#common routines for dealing with rom images and romantic comedies
from inc_noesis import *
import os

ReadRomImageFlag_None = 0
ReadRomImageFlag_Interleave = (1 << 0)
ReadRomImageFlag_EndianSwap_Word = (1 << 1)
ReadRomImageFlag_EndianSwap_DWord = (1 << 2)

def readRomImages(imageBasePath, imageSet, imageSize, flags = ReadRomImageFlag_None, printStatus = False):
	imageCount = len(imageSet)
	data = bytearray()
	for imageIndex in range(0, imageCount):
		imagePath = imageSet[imageIndex]
		imageFilePath = os.path.join(imageBasePath, imagePath)
		imageFileSize = os.path.getsize(imageFilePath)
		if imageFileSize != imageSize:
			if printStatus:
				print("Bad/no image -", imageFileSize, "-", imageFilePath)
			return None
		if printStatus:
			print("Loading", imageFilePath)
		with open(imageFilePath, "rb") as f:
			data += f.read()
			
	if flags & ReadRomImageFlag_Interleave:
		data = noesis.interleaveUniformBytes(data, 1, imageSize)
	if flags & ReadRomImageFlag_EndianSwap_Word:
		data = noesis.swapEndianArray(data, 2)
	if flags & ReadRomImageFlag_EndianSwap_DWord:
		data = noesis.swapEndianArray(data, 4)
		
	return data

#snes address to file functions assume a headerless image

def snesHiRomPtrToFileOffset(addr):
	if addr == 0:
		return 0
	if addr >= 0xC00000:
		return addr - 0xC00000
	#assume we're using the other set of linear banks
	return addr - 0x400000

def snesLoRomPtrToFileOffset(addr):
	if addr == 0:
		return 0
	if addr >= 0x800000:
		addr -= 0x800000
	bank = (addr >> 16)
	if not bank & 1:
		addr -= 0x8000
	bank >>= 1
	return (bank << 16) | (addr & 0xFFFF)

def snesFileOffsetToLoRomPtr(offset, bankRebase = 0x80):
	if offset == 0:
		return 0
	bank = offset >> 16
	addr = offset & 0xFFFF
	bank <<= 1
	if addr >= 0x8000:
		bank += 1
		addr -= 0x8000

	bank += bankRebase
	addr += 0x8000
	
	return (bank << 16) | addr
	
def snesSkipReaderHeader(data):
	smcSize = len(data) & 1023
	if smcSize:
		data = data[smcSize:]
	return data
	
def snesIsLoRom(data):
	nchecskum = noeUnpack("<H", data[0x7FDC:0x7FDE])[0]
	checskum = noeUnpack("<H", data[0x7FDE:0x7FE0])[0]
	return (checskum ^ nchecskum) == 0xFFFF
	
def snesRecalcChecksum(data, isLoRom):
	#haven't tested this over a very wide variety of data, it's possible that more exceptions exist
	dataSize = len(data)
	dupSizeLut = (0x80000, 0x100000, 0x200000, 0x200000, 0x400000)
	dupChunkSize = dupSizeLut[min(dataSize >> 20, len(dupSizeLut) - 1)]
	dupSize = dataSize & (dupChunkSize - 1)
	cdata = bytearray(data) + data[dataSize - dupSize:] * ((dupChunkSize // dupSize) - 1) if dupSize > 0 else data
	chkOffset = 0x7FDC if isLoRom else 0xFFDC
	cdata[chkOffset + 0] = 0xFF
	cdata[chkOffset + 1] = 0xFF
	cdata[chkOffset + 2] = 0x00
	cdata[chkOffset + 3] = 0x00
	checksum = 0
	for offset in range(len(cdata)):
		checksum = (checksum + cdata[offset]) & 0xFFFF
	nchecksum = checksum ^ 0xFFFF
	data[chkOffset + 0] = nchecksum & 0xFF
	data[chkOffset + 1] = nchecksum >> 8
	data[chkOffset + 2] = checksum & 0xFF
	data[chkOffset + 3] = checksum >> 8

def gameboyRecalcChecksum(data):
	#header checksum
	checksum = 0
	for offset in range(0x134, 0x14D):
		checksum = (checksum - data[offset] - 1) & 0xFF
	data[0x14D] = checksum

	#global checksum
	checksum = 0
	for offset in range(0x14E):
		checksum = (checksum + data[offset]) & 0xFFFF
	for offset in range(0x150, len(data)):
		checksum = (checksum + data[offset]) & 0xFFFF
	data[0x14E] = checksum >> 8
	data[0x14F] = checksum & 0xFF

def genesisRecalcChecksum(data):
	checksum = 0
	for offset in range(0x200, len(data), 2):
		checksum = (checksum + (data[offset] << 8) + data[offset + 1]) & 0xFFFF
	data[0x18E] = checksum >> 8
	data[0x18F] = checksum & 0xFF
